@@ -29,6 +29,67 @@ logger = logging.getLogger(__name__)
 
 
 
+NOW_PLAYING_QUERY = {
+    'remote': ['remote', '?'],
+    'remote_title': [],
+    'current_title': ['current_title', '?'],
+    'remoteMeta': [],
+    'artist': ['artist', '?'],
+    'album_id': [],
+    'filesize': [],
+    'genre': ['genre', '?'],
+    'disc': [],
+    'artwork_track_id': [],
+    'comment': [],
+    'album': ['album', '?'],
+    'bpm': [],
+    'genre_id': [],
+    'disccount': [],
+    'artist_id': [],
+    'tracknum': [],
+    'tagversion': [],
+    'compilation': [],
+    'dlna_profile': [],
+    'channels': [],
+    'playcount': [],
+    'samplerate': [],
+    'can_seek': [],
+    'id': [],
+    'title': ['current_title', '?'],
+    'coverid': [],
+    'duration': ['duration', '?'],
+    'coverart': [],
+    'modificationTime': [],
+    'type': [],
+    'bitrate': [],
+    'year': [],
+    'addedTime': [],
+    'artwork_url': [],
+    'lastUpdated': [],
+    'playlist_loop': [],
+    'player_name': [],
+    'digital_volume_control': [],
+    'signalstrength': [],
+    'playlist_cur_index': [],
+    'mode': ['mode', '?'],
+    'playlist_timestamp': [],
+    'power': [],
+    'playlist_tracks': [],
+    'rate': [],
+    'player_connected': [],
+    'time': ['time', '?'],
+    'player_ip': [],
+    'seq_no': [],
+    'playlist mode': [],
+    'playlist shuffle': [],
+    'playlist repeat': [],
+    'mixer volume': []}
+
+
+
+
+
+
 class QueryLMS():
     '''Class to handle queries for an LMS player
     
@@ -54,14 +115,18 @@ class QueryLMS():
         server_query_url(str): url to use when querying host status
         server_base_url(str): base url of server: http://host:port/
         handle_requests_exceptions(bool): True: quietly handle exceptions; False: raise exceptions
+        request_timeout(int): seconds to wait for server to respond
         
     
+    Additional API documentation: https://github.com/elParaguayo/LMS-CLI-Documentation/blob/master/LMS-CLI.md
         '''
     def __init__(self, host=None, port=None, 
                  player_name=None, 
                  player_id=None, 
-                 scan_timeout=5,
-                 handle_requests_exceptions=False):
+                 scan_timeout=1,
+                 handle_requests_exceptions=False,
+                 request_timeout=5
+                ):
         '''inits QueryLMS Class with host, port, player_id, player_name and scan_timeout
         
         Args:
@@ -72,12 +137,14 @@ class QueryLMS():
             scan_timeout(int): seconds to search for LMS host
         '''
         self.handle_requests_exceptions=handle_requests_exceptions
+
         self.host = host
-        self.port = port        
-        self.player_id = player_id
-        self.player_name = player_name
+        self.port = port
+        self.request_timeout = request_timeout
         self.scan_timeout = scan_timeout
         self.set_server()
+        self.player_id = player_id
+        self.player_name = player_name
         
     
     @property
@@ -106,17 +173,8 @@ class QueryLMS():
     @player_name.setter
     def player_name(self, player_name):
         self._player_name = player_name
-#         if player_name and not self.player_id:
-#             player_id = None
-#             logging.info(f'attempting to locate player_id for {player_name}')
-#             for p in self.get_players():
-#                 if 'name' in p and 'playerid' in p:
-#                     if p['name'] == player_name:
-#                         player_id = p['playerid']
-            
-#             self.player_id = player_id
+        self._set_player()
                 
-
     @property
     def player_id(self):
         '''LMS player unique hexidecimal id (str)'''
@@ -138,7 +196,16 @@ class QueryLMS():
                 pass
         else:
             raise exception()
-        
+            
+    def _set_player(self):
+        if self.player_name:
+            for p in self.get_players():
+                if p.get('name') == self.player_name:
+                    self.player_id = p.get('playerid', '')
+                    break
+        if self.player_name and not self.player_id:
+            logging.warning(f'could not set player_id for player "{self.player_name}"')
+    
     def set_server(self):
         '''set the server details using "host" and "port"
         
@@ -185,21 +252,9 @@ class QueryLMS():
             base_url = constants.LMS_QUERY_BASE_URL.format(self.host, self.port)
             query_url = constants.LMS_QUERY_ENDPOINT.format(base_url)
                     
-#         self.lms_server = {'host': my_host, 'port': my_port}          
         self.server_base_url = base_url
         self.server_query_url = query_url
         
-        if self.player_name and not self.player_id:
-            player_id = None
-#             logging.info(f'attempting to locate player_id for {player_name}')
-            for p in self.get_players():
-                if 'name' in p and 'playerid' in p:
-                    if p['name'] == self.player_name:
-                        player_id = p['playerid']
-                        break
-            
-            self.player_id = player_id
-        
 
     
     @staticmethod
@@ -257,14 +312,17 @@ class QueryLMS():
 
     # Basic Query
     #####################################
-    def query(self, player_id="", *args):
+    def query(self, player_id=None, *args):
+        if not self.player_id:
+            player_id = self.player_id
+            
         r = {}
         retval = {}
         params = json.dumps({'id': 1, 'method': 'slim.request',
                              'params': [player_id, list(args)]})
         if self.server_query_url:
             try:
-                r = requests.post(self.server_query_url, params)
+                r = requests.post(timeout=self.request_timeout, url=self.server_query_url, data=params)
             except requests.exceptions.RequestException as e:
                 if self.handle_requests_exceptions:
                     logging.warning(f'error making connection to server: {e}')
@@ -307,7 +365,7 @@ class QueryLMS():
         return len(self.get_artists())
 
     def get_radios_count(self):
-        '''query server for total number of radios
+        '''query server for total number of saved radio stations
         
         Returns:
             (int): count of unique radios connected'''
@@ -320,6 +378,11 @@ class QueryLMS():
             (int): count of unique players connected'''
         return self.query("", "player", "count", "?")['_count']    
     
+    def get_favorite_radio(self):
+        '''return favorited radio stations'''
+        return self.query(self.player_id, 'favorites', 
+                          'items', 0, 99,)['loop_loop']    
+    
     def get_players(self):
         '''query server for connected player information
         
@@ -439,9 +502,6 @@ class QueryLMS():
 
     def play_radio(self, radio):
         '''play radio??? on associated player'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
         return self.query(self.player_id, "favorites", "playlist", "play",
                           "item_id:" + str(radio))
 
@@ -450,10 +510,6 @@ class QueryLMS():
         
         Returns:
             (dict): {}'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         return self.query(self.player_id, "pause")
 
     def skip_songs(self, amount=1):
@@ -464,10 +520,6 @@ class QueryLMS():
         
         Returns:
             (dict): {}'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-
         if amount > 0:
             amount = "+" + str(amount)
         else:
@@ -479,10 +531,6 @@ class QueryLMS():
         
         Returns:
             (dict): {}'''
-        self._check_attribute(attribute='player_id', 
-                          check_value=True, 
-                          invalid_values=[None, ''])
-
         return self.skip_songs(-1)
 
     def next_song(self):
@@ -490,10 +538,6 @@ class QueryLMS():
         
         Returns:
             (dict): {}'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         return self.skip_songs()
     
     def get_volume(self):
@@ -501,10 +545,6 @@ class QueryLMS():
         
         Returns:
             (str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         volume = self.query(self.player_id, "mixer", "volume", "?")
         if len(volume):
             volume = volume['_volume']
@@ -520,10 +560,6 @@ class QueryLMS():
             
         Returns:
             (dict): {}'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         return self.query(self.player_id, "mixer", "volume", volume)
         
 
@@ -532,102 +568,64 @@ class QueryLMS():
         
         Returns:
             (str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         title = self.query(self.player_id, "current_title", "?")
-        if len(title):
-            title = title['_current_title']
-        else:
-            title = ""
-        return title
-
+        
+        return title.get('_current_title', '')
+    
     def get_current_artist(self):
         '''query associated player for currently playing artist
         
         Returns:
             (str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-
         artist = self.query(self.player_id, "artist", "?")
-        if len(artist):
-            artist = artist['_artist']
-        else:
-            artist = ""
-        return artist
+        return artist.get('_artist', '')    
 
     def get_current_album(self):
         '''query associated player for currently playing track album
         
         Returns:
             (str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
         
         album = self.query(self.player_id, "album", "?")
-        if len(album):
-            album = album['_album']
-        else:
-            album = ""
-        return album
+        return album.get('_album', '')
 
     def get_current_title(self):
         '''query associated player for currently playing track title
         
         Returns:
             (str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
         
         title = self.query(self.player_id, "title", "?")
-        if len(title):
-            title = title['_title']
-        else:
-            title = ""
-        return title
-
+        return title.get('_title', '')
+    
     def get_current_radio_title(self, radio):
-        '''???'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
+        '''return title of favorite radio stations'''
         
-        return self.query(self.player_id, "favorites",
-                          "items", 0, 99)['loop_loop'][radio]['name']
-
+        title = self.query(self.player_id, 'favorites', 'items', 0, 99)
+        
+        loop = title.get('loop_loop', {})
+        
+        name = loop[radio].get('name', '')
+        
+        return name
+    
+    @property
     def is_playing_remote_stream(self):
-        '''???'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
-        return self.query(self.player_id, "remote", "?")['_remote'] == 1
+        remote = self.query(self.player_id, "remote", "?")        
+        return remote.get('_remote', False)
 
     def get_artist_album(self, artist_id):
         '''query associated player for currently playing album artist
         
         Returns:
             (str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
-        return self.query(self.player_id, "albums", 0, 99, "tags:al",
-                          "artist_id:" + str(artist_id))['albums_loop']
+        album_artist = self.query(self.player_id, 'albums', 0, 99, 'tags:al', f'artist_id:{artist_id}')
+        return album_artist.get('albums_loop', '')
 
     def get_alarms(self, enabled=True):
         '''???'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         if enabled:
-            alarmsEnabled = self.get_player_pref(self.player_id, "alarmsEnabled")
+            alarmsEnabled = self.get_player_pref("alarmsEnabled")
             if alarmsEnabled == "0":
                 return {}
             alarm_filter = "enabled"
@@ -668,72 +666,144 @@ class QueryLMS():
         else:
             return {"alarmtime": alarmtime.seconds, "delta": delta.seconds}
 
-    def get_now_playing(self):
-        '''query associated player for now playing information including:
-        * album
-        * artist
-        * artwork_url
-        * duration
-        * genre
-        * coverid
-        * id
-        * title'''  
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        status_keys = ['time', 'mode']
+#     def get_now_playing(self):
+#         '''query associated player for now playing information including:
+#         * album
+#         * artist
+#         * artwork_url
+#         * duration
+#         * genre
+#         * coverid
+#         * id
+#         * title'''
+# #         pass
+#         self._check_attribute(attribute='player_id', 
+#                               check_value=True, 
+#                               invalid_values=[None, ''])
+#         status_keys = ['time', 'mode']
         
-        status = self.query(self.player_id, 'status')
-        track_id = None
-        song_info = None
-        now_playing_info = {}
+#         status = self.query(self.player_id, 'status')
+#         track_id = None
+#         song_info = None
+#         now_playing_info = {}
         
-        if status:
-            try:
-                playing_track = self.query(self.player_id, 'status', 
-                                           int(status['playlist_cur_index']), 1, '-')['playlist_loop'][0]
-                track_id = playing_track['id']
-                song_info = self.query('', 'songinfo', 0, 100, 'track_id:'+str(track_id), 'tags:a,c,d,e,g,l')['songinfo_loop']
-                for key in status_keys:
-                    if key in status:
-                        now_playing_info[key] = status[key]
-                    else:
-                        now_playing_info[key] = None
-            except (KeyError, IndexError):
-                pass
+#         if status:
+#             try:
+#                 playing_track = self.query(self.player_id, 'status', 
+#                                            int(status['playlist_cur_index']), 1, '-')['playlist_loop'][0]
+#                 track_id = playing_track['id']
+#                 song_info = self.query('', 'songinfo', 0, 100, 'track_id:'+str(track_id), 'tags:a,c,d,e,g,l')['songinfo_loop']
+#                 for key in status_keys:
+#                     if key in status:
+#                         now_playing_info[key] = status[key]
+#                     else:
+#                         now_playing_info[key] = None
+#             except (KeyError, IndexError):
+#                 pass
        
-        if song_info:
-            try:
-                for each in song_info:
-                    for key in each:
-                        now_playing_info[key] = each[key]
-                coverid = 0
-                if 'coverid' in now_playing_info:
-                    if now_playing_info['coverid'].startswith('-'):
-                        pass
-                    else:
-                        coverid = now_playing_info['coverid']
-                now_playing_info['artwork_url'] = f'{self.server_base_url}music/{coverid}/cover.jpg'
-            except (KeyError, IndexError):
-                pass
+#         if song_info:
+#             try:
+#                 for each in song_info:
+#                     for key in each:
+#                         now_playing_info[key] = each[key]
+#                 coverid = 0
+#                 if 'coverid' in now_playing_info:
+#                     if now_playing_info['coverid'].startswith('-'):
+#                         pass
+#                     else:
+#                         coverid = now_playing_info['coverid']
+#                 now_playing_info['artwork_url'] = f'{self.server_base_url}music/{coverid}/cover.jpg'
+#             except (KeyError, IndexError):
+#                 pass
             
 
-        return now_playing_info
+#         return now_playing_info
     
-    def get_player_pref(self, pref):
-        '''???'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
+    def get_now_playing(self):
+        status = self.query(self.player_id, 'status', '-')
+        playlist = status.get('playlist_loop', [])
         
+        try:
+            playing_track = playlist[0]
+        except IndexError:
+            logging.warning('no valid playlist was returned')
+            playing_track = {}
+        
+        track_id = playing_track.get('id', 0)
+    
+        track_info = self.query(self.player_id, 'songinfo', '-', 100, f'track_id:{track_id}')
+        
+        info_list = track_info.get('songinfo_loop', [])
+        
+        now_playing = {}
+        for i in info_list:
+            for k, v in i.items():
+                now_playing[k] = v
+        
+        coverid = now_playing.get('coverid', None)
+        
+        if coverid:
+            artwork_url = f'{self.server_base_url}music/{coverid}/cover.jpg'
+        else:
+            artwork_url = ''
+        
+        now_playing['artwork_url'] = artwork_url
+        now_playing = {**now_playing, **status}
+        
+        # first run - try to populate missing keys
+        now_playing = self._add_keys(now_playing)
+        # fill in null values for remaining keys
+        now_playing = self._add_keys(now_playing, True)
+        
+        if now_playing.get('remote'):
+            now_playing['title'] = now_playing.get('remoteMeta', now_playing.get('title', '')).get('title', '')
+        
+        # ensure there is always an album_id value
+        if not now_playing.get('album_id', False):
+            now_playing['album_id'] = 'no_album_id'
+        
+        return now_playing
+    
+    
+    def _add_keys(self, now_playing, add_blank=False):
+        '''fill in missing keys using the NOW_PLAYING_QUERY constant
+        
+        Run with add_blank=False to use queries stored in NOW_PLAYING to attempt to fill
+        in missing key/value pairs.
+        
+        Run with add_blak=True to add blank strings to ensure that all values are populated
+        
+        Args:
+            now_playin(dict): dictionary of now playing values
+            add_blank(bool): True fill in any missing values with a '' string
+        
+        '''
+        for k, query in NOW_PLAYING_QUERY.items():
+            result = {}
+            if not now_playing.get(k, False):
+                if add_blank:
+                    now_playing[k] = ''
+                    continue
+                else:
+                    logging.debug('adding missing keys')
+
+
+                if query:
+                    result = self.query(self.player_id, *query)
+                for i, j in result.items():
+                    if i.strip('_') in NOW_PLAYING_QUERY.keys():
+                        logging.debug(f'adding "{k}: {j}" to now_playing')
+                        now_playing[i.strip('_')] = j
+                        
+        return now_playing              
+            
+
+    def get_player_pref(self, pref):
+        '''???'''      
         return self.query(self.player_id, "playerpref", pref, "?")['_p2']
 
     def set_player_pref(self, pref, value):
         '''???'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         self.query(self.player_id, "playerpref", pref, value)
 
     def display(self, line1, line2, duration=5):
@@ -742,10 +812,6 @@ class QueryLMS():
         Args:
             line1(str)
             line1(str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-        
         self.query(self.player_id, "display", line1, line2, duration)
 
     def display_all(self, line1, line2, duration=5):
@@ -754,10 +820,6 @@ class QueryLMS():
         Args:
             line1(str)
             line1(str)'''
-        self._check_attribute(attribute='player_id', 
-                              check_value=True, 
-                              invalid_values=[None, ''])
-
         players = self.get_players()
         for player in players:
             self.display(player['playerid'], line1, line2, duration)
@@ -765,10 +827,3 @@ class QueryLMS():
 
 
 
-
-
-
-
-
-
-
